@@ -10,6 +10,8 @@
 #include <OgreCompositorManager.h>
 
 #include <opencv2/calib3d.hpp>
+#include <opencv2/core/utils/configuration.private.hpp>
+
 
 namespace cv
 {
@@ -21,7 +23,7 @@ const char* RESOURCEGROUP_NAME = "OVIS";
 Ptr<Application> _app;
 
 static const char* RENDERSYSTEM_NAME = "OpenGL 3+ Rendering Subsystem";
-static std::vector<String> _extraResourceLocations;
+static std::set<String> _extraResourceLocations;
 
 // convert from OpenCV to Ogre coordinates:
 static Quaternion toOGRE(Degree(180), Vector3::UNIT_X);
@@ -31,27 +33,47 @@ WindowScene::~WindowScene() {}
 
 void _createTexture(const String& name, Mat image)
 {
+    PixelFormat format;
+    switch(image.type())
+    {
+    case CV_8UC4:
+        format = PF_BYTE_BGRA;
+        break;
+    case CV_8UC3:
+        format = PF_BYTE_BGR;
+        break;
+    case CV_8UC1:
+        format = PF_BYTE_L;
+        break;
+    case CV_16UC1:
+        format = PF_L16;
+        break;
+    case CV_32FC1:
+        format = PF_FLOAT32_R;
+        break;
+    default:
+        CV_Error(Error::StsBadArg, "currently supported formats are only CV_8UC1, CV_8UC3, CV_8UC4, CV_16UC1, CV_32FC1");
+        break;
+    }
+
     TextureManager& texMgr = TextureManager::getSingleton();
     TexturePtr tex = texMgr.getByName(name, RESOURCEGROUP_NAME);
 
-    Image im;
-    im.loadDynamicImage(image.ptr(), image.cols, image.rows, 1, PF_BYTE_BGR);
-
-    if (tex)
+    if(!tex)
     {
-        // update
-        PixelBox box = im.getPixelBox();
-        tex->getBuffer()->blitFromMemory(box, box);
-        return;
+        tex = texMgr.createManual(name, RESOURCEGROUP_NAME, TEX_TYPE_2D, image.cols, image.rows,
+                                  MIP_DEFAULT, format);
     }
 
-    texMgr.loadImage(name, RESOURCEGROUP_NAME, im);
+    PixelBox box(image.cols, image.rows, 1, format, image.ptr());
+    box.rowPitch = image.step[0] / PixelUtil::getNumElemBytes(format);
+    tex->getBuffer()->blitFromMemory(box);
 }
 
 static void _convertRT(InputArray rot, InputArray tvec, Quaternion& q, Vector3& t, bool invert = false)
 {
-    CV_Assert(rot.empty() || rot.rows() == 3 || rot.size() == Size(3, 3),
-              tvec.empty() || tvec.rows() == 3);
+    CV_Assert_N(rot.empty() || rot.rows() == 3 || rot.size() == Size(3, 3),
+                tvec.empty() || tvec.rows() == 3);
 
     q = Quaternion::IDENTITY;
     t = Vector3::ZERO;
@@ -98,11 +120,11 @@ static void _setCameraIntrinsics(Camera* cam, InputArray _K, const Size& imsize)
 
     Matx33f K = _K.getMat();
 
-    float near = cam->getNearClipDistance();
-    float top = near * K(1, 2) / K(1, 1);
-    float left = -near * K(0, 2) / K(0, 0);
-    float right = near * (imsize.width - K(0, 2)) / K(0, 0);
-    float bottom = -near * (imsize.height - K(1, 2)) / K(1, 1);
+    float zNear = cam->getNearClipDistance();
+    float top = zNear * K(1, 2) / K(1, 1);
+    float left = -zNear * K(0, 2) / K(0, 0);
+    float right = zNear * (imsize.width - K(0, 2)) / K(0, 0);
+    float bottom = -zNear * (imsize.height - K(1, 2)) / K(1, 1);
 
     // use frustum extents instead of setFrustumOffset as the latter
     // assumes centered FOV, which is not the case
@@ -125,7 +147,7 @@ static SceneNode& _getSceneNode(SceneManager* sceneMgr, const String& name)
         if(mo)
             return *mo->getParentSceneNode()->getParentSceneNode();
     }
-    catch (ItemIdentityException&)
+    catch (const ItemIdentityException&)
     {
         // ignore
     }
@@ -135,7 +157,7 @@ static SceneNode& _getSceneNode(SceneManager* sceneMgr, const String& name)
         if (!mo)
             mo = sceneMgr->getMovableObject(name, "Light");
     }
-    catch (ItemIdentityException&)
+    catch (const ItemIdentityException&)
     {
         // ignore
     }
@@ -146,10 +168,20 @@ static SceneNode& _getSceneNode(SceneManager* sceneMgr, const String& name)
     return *mo->getParentSceneNode();
 }
 
+static ColourValue convertColor(const Scalar& val)
+{
+    // BGR 0..255 (uchar) to RGB 0..1
+    ColourValue ret = ColourValue(val[2], val[1], val[0]) / 255;
+    ret.saturate();
+    return ret;
+}
+
+class WindowSceneImpl;
+
 struct Application : public OgreBites::ApplicationContext, public OgreBites::InputListener
 {
     Ptr<LogManager> logMgr;
-    Ogre::SceneManager* sceneMgr;
+    WindowSceneImpl* mainWin;
     Ogre::String title;
     uint32_t w;
     uint32_t h;
@@ -157,9 +189,13 @@ struct Application : public OgreBites::ApplicationContext, public OgreBites::Inp
     int flags;
 
     Application(const Ogre::String& _title, const Size& sz, int _flags)
-        : OgreBites::ApplicationContext("ovis", false), sceneMgr(NULL), title(_title), w(sz.width),
+        : OgreBites::ApplicationContext("ovis"), mainWin(NULL), title(_title), w(sz.width),
           h(sz.height), key_pressed(-1), flags(_flags)
     {
+        if(utils::getConfigurationParameterBool("OPENCV_OVIS_VERBOSE_LOG", false))
+            return;
+
+        // set default log with low log level
         logMgr.reset(new LogManager());
         logMgr->createLog("ovis.log", true, true, true);
         logMgr->setLogDetail(LL_LOW);
@@ -170,25 +206,26 @@ struct Application : public OgreBites::ApplicationContext, public OgreBites::Inp
         // empty impl to show cursor
     }
 
-    bool keyPressed(const OgreBites::KeyboardEvent& evt)
+    bool keyPressed(const OgreBites::KeyboardEvent& evt) CV_OVERRIDE
     {
         key_pressed = evt.keysym.sym;
         return true;
     }
 
-    bool oneTimeConfig()
+    bool oneTimeConfig() CV_OVERRIDE
     {
-        Ogre::RenderSystem* rs = getRoot()->getRenderSystemByName(RENDERSYSTEM_NAME);
-        CV_Assert(rs);
+        Ogre::String rsname = utils::getConfigurationParameterString("OPENCV_OVIS_RENDERSYSTEM", RENDERSYSTEM_NAME);
+        Ogre::RenderSystem* rs = getRoot()->getRenderSystemByName(rsname);
+        CV_Assert(rs && "Could not find rendersystem");
         getRoot()->setRenderSystem(rs);
         return true;
     }
 
     OgreBites::NativeWindowPair createWindow(const Ogre::String& name, uint32_t _w, uint32_t _h,
-                                             NameValuePairList miscParams = NameValuePairList())
+                                             NameValuePairList miscParams = NameValuePairList()) CV_OVERRIDE
     {
         Ogre::String _name = name;
-        if (!sceneMgr)
+        if (!mainWin)
         {
             _w = w;
             _h = h;
@@ -198,7 +235,8 @@ struct Application : public OgreBites::ApplicationContext, public OgreBites::Inp
         if (flags & SCENE_AA)
             miscParams["FSAA"] = "4";
 
-        miscParams["vsync"] = "true";
+        miscParams["vsync"] = Ogre::StringConverter::toString(
+            !utils::getConfigurationParameterBool("OPENCV_OVIS_NOVSYNC", false));
 
         OgreBites::NativeWindowPair ret =
             OgreBites::ApplicationContext::createWindow(_name, _w, _h, miscParams);
@@ -207,16 +245,31 @@ struct Application : public OgreBites::ApplicationContext, public OgreBites::Inp
         return ret;
     }
 
-    void locateResources()
+#if OGRE_VERSION < ((1 << 16) | (12 << 8) | 3)
+    void destroyWindow(const Ogre::String& name)
+    {
+        for (auto it = mWindows.begin(); it != mWindows.end(); ++it)
+        {
+            if (it->render->getName() != name)
+                continue;
+            mRoot->destroyRenderTarget(it->render);
+            mWindows.erase(it);
+            return;
+        }
+    }
+#endif
+
+    size_t numWindows() const { return mWindows.size(); }
+
+    void locateResources() CV_OVERRIDE
     {
         OgreBites::ApplicationContext::locateResources();
         ResourceGroupManager& rgm = ResourceGroupManager::getSingleton();
         rgm.createResourceGroup(RESOURCEGROUP_NAME);
 
-        for (size_t i = 0; i < _extraResourceLocations.size(); i++)
+        for (auto loc :_extraResourceLocations)
         {
-            String loc = _extraResourceLocations[i];
-            String type = StringUtil::endsWith(loc, ".zip") ? "Zip" : "FileSystem";
+            const char* type = StringUtil::endsWith(loc, ".zip") ? "Zip" : "FileSystem";
 
             if (!FileSystemLayer::fileExists(loc))
             {
@@ -227,7 +280,7 @@ struct Application : public OgreBites::ApplicationContext, public OgreBites::Inp
         }
     }
 
-    void setup()
+    void setup() CV_OVERRIDE
     {
         OgreBites::ApplicationContext::setup();
 
@@ -243,22 +296,23 @@ class WindowSceneImpl : public WindowScene
     Root* root;
     SceneManager* sceneMgr;
     SceneNode* camNode;
-    RenderWindow* rWin;
+    RenderTarget* rWin;
     Ptr<OgreBites::CameraMan> camman;
     Ptr<Rectangle2D> bgplane;
+    std::unordered_map<AnimationState*, Controller<Real>*> frameCtrlrs;
 
-    Ogre::RenderTarget* frameSrc;
     Ogre::RenderTarget* depthRTT;
+    int flags;
 public:
-    WindowSceneImpl(Ptr<Application> app, const String& _title, const Size& sz, int flags)
-        : title(_title), root(app->getRoot()), depthRTT(NULL)
+    WindowSceneImpl(Ptr<Application> app, const String& _title, const Size& sz, int _flags)
+        : title(_title), root(app->getRoot()), depthRTT(NULL), flags(_flags)
     {
-        if (!app->sceneMgr)
+        if (!app->mainWin)
         {
-            flags |= SCENE_SEPERATE;
+            flags |= SCENE_SEPARATE;
         }
 
-        if (flags & SCENE_SEPERATE)
+        if (flags & SCENE_SEPARATE)
         {
             sceneMgr = root->createSceneManager("DefaultSceneManager", title);
             RTShader::ShaderGenerator& shadergen = RTShader::ShaderGenerator::getSingleton();
@@ -269,7 +323,8 @@ public:
         }
         else
         {
-            sceneMgr = app->sceneMgr;
+            sceneMgr = app->mainWin->sceneMgr;
+            bgplane = app->mainWin->bgplane;
         }
 
         if(flags & SCENE_SHOW_CS_CROSS)
@@ -288,16 +343,27 @@ public:
         {
             camman.reset(new OgreBites::CameraMan(camNode));
             camman->setStyle(OgreBites::CS_ORBIT);
-            camNode->setFixedYawAxis(true, Vector3::NEGATIVE_UNIT_Y);
+            camman->setFixedYaw(false);
         }
 
-        if (!app->sceneMgr)
+        if (!app->mainWin)
         {
-            app->sceneMgr = sceneMgr;
+            CV_Assert((flags & SCENE_OFFSCREEN) == 0 && "off-screen rendering for main window not supported");
+
+            app->mainWin = this;
             rWin = app->getRenderWindow();
             if (camman)
                 app->addInputListener(camman.get());
         }
+        else if (flags & SCENE_OFFSCREEN)
+        {
+            // render into an offscreen texture
+            TexturePtr tex = TextureManager::getSingleton().createManual(
+                title, RESOURCEGROUP_NAME, TEX_TYPE_2D, sz.width, sz.height, 0, PF_BYTE_RGB,
+                TU_RENDERTARGET, NULL, false, flags & SCENE_AA ? 4 : 0);
+            rWin = tex->getBuffer()->getRenderTarget();
+            rWin->setAutoUpdated(false); // only update when requested
+        }
         else
         {
             OgreBites::NativeWindowPair nwin = app->createWindow(title, sz.width, sz.height);
@@ -307,69 +373,156 @@ public:
         }
 
         rWin->addViewport(cam);
-        frameSrc = rWin;
+    }
 
-        if (flags & SCENE_RENDER_FLOAT)
+    ~WindowSceneImpl()
+    {
+        if (flags & SCENE_SEPARATE)
         {
-            // also render into an offscreen texture
-            // currently this draws everything twice, but we spare the float->byte conversion for display
-            TexturePtr tex = TextureManager::getSingleton().createManual(
-                title + "_rt", RESOURCEGROUP_NAME, TEX_TYPE_2D, sz.width, sz.height, 0, PF_FLOAT32_RGBA,
-                TU_RENDERTARGET);
-            frameSrc = tex->getBuffer()->getRenderTarget();
-            frameSrc->addViewport(cam);
+            TextureManager& texMgr =  TextureManager::getSingleton();
+
+            MaterialManager::getSingleton().remove(bgplane->getMaterial());
+            bgplane.release();
+            String texName = "_"+sceneMgr->getName() + "_DefaultBackground";
+            texMgr.remove(texName, RESOURCEGROUP_NAME);
+
+            texName = sceneMgr->getName() + "_Background";
+            if(texMgr.resourceExists(texName, RESOURCEGROUP_NAME))
+            {
+                texMgr.remove(texName, RESOURCEGROUP_NAME);
+            }
+        }
+
+        if(_app->mainWin == this && (flags & SCENE_SEPARATE))
+        {
+            // this is the root window owning the context
+            CV_Assert(_app->numWindows() == 1 && "the first OVIS window must be deleted last");
+            _app->closeApp();
+            _app.release();
+        }
+        else if (flags & SCENE_OFFSCREEN)
+        {
+            TextureManager::getSingleton().remove(title, RESOURCEGROUP_NAME);
+        }
+        else
+        {
+            _app->destroyWindow(title);
         }
     }
 
-    void setBackground(InputArray image)
+    void setBackground(InputArray image) CV_OVERRIDE
     {
-        CV_Assert(image.type() == CV_8UC3, bgplane);
+        CV_Assert(bgplane);
 
         String name = sceneMgr->getName() + "_Background";
 
         _createTexture(name, image.getMat());
 
-        // correct for pixel centers
-        Vector2 pc(0.5 / image.cols(), 0.5 / image.rows());
-        bgplane->setUVs(pc, Vector2(pc[0], 1 - pc[1]), Vector2(1 - pc[0], pc[1]), Vector2(1, 1) - pc);
+        bgplane->setDefaultUVs();
 
         Pass* rpass = bgplane->getMaterial()->getBestTechnique()->getPasses()[0];
         rpass->getTextureUnitStates()[0]->setTextureName(name);
+        rpass->getTextureUnitStates()[0]->setTextureAddressingMode(TAM_CLAMP);
 
         // ensure bgplane is visible
         bgplane->setVisible(true);
     }
 
-    void setCompositors(const std::vector<String>& names)
+    void setCompositors(const std::vector<String>& names) CV_OVERRIDE
     {
-        Viewport* vp = frameSrc->getViewport(0);
         CompositorManager& cm = CompositorManager::getSingleton();
+        // this should be applied to all owned render targets
+        Ogre::RenderTarget* targets[] = {rWin, depthRTT};
 
-        cm.removeCompositorChain(vp); // remove previous configuration
-
-        for(size_t i = 0; i < names.size(); i++)
+        for(int j = 0; j < 2; j++)
         {
-            if (!cm.addCompositor(vp, names[i])) {
-                LogManager::getSingleton().logError("Failed to add compositor: " + names[i]);
-                continue;
+            Ogre::RenderTarget* tgt = targets[j];
+            if(!tgt) continue;
+
+            Viewport* vp = tgt->getViewport(0);
+            cm.removeCompositorChain(vp); // remove previous configuration
+
+            for(size_t i = 0; i < names.size(); i++)
+            {
+                if (!cm.addCompositor(vp, names[i])) {
+                    LogManager::getSingleton().logError("Failed to add compositor: " + names[i]);
+                    continue;
+                }
+                cm.setCompositorEnabled(vp, names[i], true);
             }
-            cm.setCompositorEnabled(vp, names[i], true);
         }
     }
 
-    void setBackground(const Scalar& color)
+    void getCompositorTexture(const String& compname, const String& texname, OutputArray out,
+                              int mrtIndex) CV_OVERRIDE
+    {
+        CompositorManager& cm = CompositorManager::getSingleton();
+        CompositorChain* chain = cm.getCompositorChain(rWin->getViewport(0));
+        CV_Assert(chain && "no active compositors");
+
+        CompositorInstance* inst = chain->getCompositor(compname);
+        if(!inst)
+            CV_Error_(Error::StsBadArg, ("no active compositor named: %s", compname.c_str()));
+
+        TexturePtr tex = inst->getTextureInstance(texname, mrtIndex);
+        if(!tex)
+            CV_Error_(Error::StsBadArg, ("no texture named: %s", texname.c_str()));
+
+        PixelFormat src_type = tex->getFormat();
+        int dst_type;
+        switch(src_type)
+        {
+        case PF_R8:
+        case PF_L8:
+            dst_type = CV_8U;
+            break;
+        case PF_BYTE_RGB:
+            dst_type = CV_8UC3;
+            break;
+        case PF_BYTE_RGBA:
+            dst_type = CV_8UC4;
+            break;
+#if OGRE_VERSION >= ((1 << 16) | (12 << 8) | 3)
+        case PF_DEPTH32F:
+#endif
+        case PF_FLOAT32_R:
+            dst_type = CV_32F;
+            break;
+        case PF_FLOAT32_RGB:
+            dst_type = CV_32FC3;
+            break;
+        case PF_FLOAT32_RGBA:
+            dst_type = CV_32FC4;
+            break;
+        case PF_L16:
+        case PF_DEPTH:
+            dst_type = CV_16U;
+            break;
+        default:
+            CV_Error(Error::StsNotImplemented, "unsupported texture format");
+        }
+
+        out.create(tex->getHeight(), tex->getWidth(), dst_type);
+
+        Mat mat = out.getMat();
+        PixelBox pb(tex->getWidth(), tex->getHeight(), 1, src_type, mat.ptr());
+        tex->getBuffer()->blitToMemory(pb, pb);
+
+        if(CV_MAT_CN(dst_type) < 3)
+            return;
+
+        // convert to OpenCV channel order
+        cvtColor(mat, mat, CV_MAT_CN(dst_type) == 3 ? COLOR_RGB2BGR : COLOR_RGBA2BGRA);
+    }
+
+    void setBackground(const Scalar& color) CV_OVERRIDE
     {
         // hide background plane
         bgplane->setVisible(false);
-
-        // BGRA as uchar
-        ColourValue _color = ColourValue(color[2], color[1], color[0], color[3]) / 255;
-        rWin->getViewport(0)->setBackgroundColour(_color);
-        if(frameSrc != rWin)
-            frameSrc->getViewport(0)->setBackgroundColour(_color);
+        rWin->getViewport(0)->setBackgroundColour(convertColor(color));
     }
 
-    void createEntity(const String& name, const String& meshname, InputArray tvec, InputArray rot)
+    void createEntity(const String& name, const String& meshname, InputArray tvec, InputArray rot) CV_OVERRIDE
     {
         Entity* ent = sceneMgr->createEntity(name, meshname, RESOURCEGROUP_NAME);
 
@@ -380,15 +533,21 @@ public:
         node->attachObject(ent);
     }
 
-    void removeEntity(const String& name) {
+    void removeEntity(const String& name) CV_OVERRIDE
+    {
         SceneNode& node = _getSceneNode(sceneMgr, name);
         node.getAttachedObject(name)->detachFromParent();
+
+        // only one of the following will do something
+        sceneMgr->destroyLight(name);
         sceneMgr->destroyEntity(name);
+        sceneMgr->destroyCamera(name);
+
         sceneMgr->destroySceneNode(&node);
     }
 
     Rect2d createCameraEntity(const String& name, InputArray K, const Size& imsize, float zFar,
-                              InputArray tvec, InputArray rot)
+                              InputArray tvec, InputArray rot) CV_OVERRIDE
     {
         MaterialPtr mat = MaterialManager::getSingleton().create(name, RESOURCEGROUP_NAME);
         Pass* rpass = mat->getTechniques()[0]->getPasses()[0];
@@ -421,10 +580,9 @@ public:
     }
 
     void createLightEntity(const String& name, InputArray tvec, InputArray rot, const Scalar& diffuseColour,
-                           const Scalar& specularColour)
+                           const Scalar& specularColour) CV_OVERRIDE
     {
         Light* light = sceneMgr->createLight(name);
-        light->setDirection(Vector3::NEGATIVE_UNIT_Z);
         // convert to BGR
         light->setDiffuseColour(ColourValue(diffuseColour[2], diffuseColour[1], diffuseColour[0]));
         light->setSpecularColour(ColourValue(specularColour[2], specularColour[1], specularColour[0]));
@@ -436,7 +594,7 @@ public:
         node->attachObject(light);
     }
 
-    void updateEntityPose(const String& name, InputArray tvec, InputArray rot)
+    void updateEntityPose(const String& name, InputArray tvec, InputArray rot) CV_OVERRIDE
     {
         SceneNode& node = _getSceneNode(sceneMgr, name);
         Quaternion q;
@@ -446,7 +604,7 @@ public:
         node.translate(t, Ogre::Node::TS_LOCAL);
     }
 
-    void setEntityPose(const String& name, InputArray tvec, InputArray rot, bool invert)
+    void setEntityPose(const String& name, InputArray tvec, InputArray rot, bool invert) CV_OVERRIDE
     {
         SceneNode& node = _getSceneNode(sceneMgr, name);
         Quaternion q;
@@ -456,7 +614,83 @@ public:
         node.setPosition(t);
     }
 
-    void setEntityProperty(const String& name, int prop, const String& value)
+	void getEntityPose(const String& name ,OutputArray R, OutputArray tvec, bool invert) CV_OVERRIDE
+    {
+		SceneNode* node = sceneMgr->getEntity(name)->getParentSceneNode();
+        Matrix3 _R;
+        // toOGRE.Inverse() == toOGRE
+        (node->getOrientation()*toOGRE).ToRotationMatrix(_R);
+
+        if (invert)
+        {
+            _R = _R.Transpose();
+        }
+
+        if (tvec.needed())
+        {
+            Vector3 _tvec = node->getPosition();
+
+            if (invert)
+            {
+                _tvec = _R * -_tvec;
+            }
+
+            Mat_<Real>(3, 1, _tvec.ptr()).copyTo(tvec);
+        }
+
+        if (R.needed())
+        {
+            Mat_<Real>(3, 3, _R[0]).copyTo(R);
+        }
+    }
+
+    void getEntityAnimations(const String& name, std::vector<String>& out) CV_OVERRIDE
+    {
+        SceneNode& node = _getSceneNode(sceneMgr, name);
+        Entity* ent = dynamic_cast<Entity*>(node.getAttachedObject(name));
+        CV_Assert(ent && "invalid entity");
+        for (auto const& anim : ent->getAllAnimationStates()->getAnimationStates())
+        {
+            out.push_back(anim.first);
+        }
+    }
+
+    void playEntityAnimation(const String& name, const String& animname, bool loop = true) CV_OVERRIDE
+    {
+        SceneNode& node = _getSceneNode(sceneMgr, name);
+        Entity* ent = dynamic_cast<Entity*>(node.getAttachedObject(name));
+        CV_Assert(ent && "invalid entity");
+        AnimationState* animstate = ent->getAnimationState(animname);
+
+        animstate->setTimePosition(0);
+        animstate->setEnabled(true);
+        animstate->setLoop(loop);
+
+        if (frameCtrlrs.find(animstate) != frameCtrlrs.end()) return;
+        frameCtrlrs.insert({
+            animstate,
+            Ogre::ControllerManager::getSingleton().createFrameTimePassthroughController(
+                Ogre::AnimationStateControllerValue::create(animstate, true)
+            )
+        });
+    }
+
+    void stopEntityAnimation(const String& name, const String& animname) CV_OVERRIDE
+    {
+        SceneNode& node = _getSceneNode(sceneMgr, name);
+        Entity* ent = dynamic_cast<Entity*>(node.getAttachedObject(name));
+        CV_Assert(ent && "invalid entity");
+        AnimationState* animstate = ent->getAnimationState(animname);
+
+        if (!animstate->getEnabled()) return;
+
+        animstate->setEnabled(false);
+        animstate->setTimePosition(0);
+        Ogre::ControllerManager::getSingleton().destroyController(frameCtrlrs[animstate]);
+        frameCtrlrs.erase(animstate);
+    }
+
+    void setEntityProperty(const String& name, int prop, const String& value, int subEntityIdx) CV_OVERRIDE
     {
         CV_Assert(prop == ENTITY_MATERIAL);
         SceneNode& node = _getSceneNode(sceneMgr, name);
@@ -467,20 +701,70 @@ public:
         Camera* cam = dynamic_cast<Camera*>(node.getAttachedObject(name));
         if(cam)
         {
+            CV_Assert(subEntityIdx == -1 && "Camera Entities do not have SubEntities");
             cam->setMaterial(mat);
             return;
         }
 
         Entity* ent = dynamic_cast<Entity*>(node.getAttachedObject(name));
         CV_Assert(ent && "invalid entity");
-        ent->setMaterial(mat);
+
+        if (subEntityIdx < 0)
+            ent->setMaterial(mat);
+        else
+            ent->getSubEntities()[subEntityIdx]->setMaterial(mat);
     }
 
-    void setEntityProperty(const String& name, int prop, const Scalar& value)
+    void setEntityProperty(const String& name, int prop, const Scalar& value) CV_OVERRIDE
     {
-        CV_Assert(prop == ENTITY_SCALE);
         SceneNode& node = _getSceneNode(sceneMgr, name);
-        node.setScale(value[0], value[1], value[2]);
+        switch(prop)
+        {
+        case ENTITY_SCALE:
+        {
+            node.setScale(value[0], value[1], value[2]);
+            break;
+        }
+        case ENTITY_ANIMBLEND_MODE:
+        {
+            Entity* ent = dynamic_cast<Entity*>(node.getAttachedObject(name));
+            CV_Assert(ent && "invalid entity");
+
+            ent->getSkeleton()->setBlendMode(static_cast<Ogre::SkeletonAnimationBlendMode>(int(value[0])));
+            break;
+        }
+        default:
+            CV_Error(Error::StsBadArg, "unsupported property");
+        }
+    }
+
+    void getEntityProperty(const String& name, int prop, OutputArray value) CV_OVERRIDE
+    {
+        SceneNode& node = _getSceneNode(sceneMgr, name);
+        switch(prop)
+        {
+        case ENTITY_SCALE:
+        {
+            Vector3 s = node.getScale();
+            Mat_<Real>(1, 3, s.ptr()).copyTo(value);
+            return;
+        }
+        case ENTITY_AABB_WORLD:
+        {
+            Entity* ent = dynamic_cast<Entity*>(node.getAttachedObject(name));
+            CV_Assert(ent && "invalid entity");
+            AxisAlignedBox aabb = ent->getWorldBoundingBox(true);
+            Vector3 mn = aabb.getMinimum();
+            Vector3 mx = aabb.getMaximum();
+            Mat_<Real> ret(2, 3);
+            Mat_<Real>(1, 3, mn.ptr()).copyTo(ret.row(0));
+            Mat_<Real>(1, 3, mx.ptr()).copyTo(ret.row(1));
+            ret.copyTo(value);
+            return;
+        }
+        default:
+            CV_Error(Error::StsBadArg, "unsupported property");
+        }
     }
 
     void _createBackground()
@@ -500,7 +784,7 @@ public:
         bgplane.reset(new Rectangle2D(true));
         bgplane->setCorners(-1.0, 1.0, 1.0, -1.0);
 
-        // correct for pixel centers
+        // use pixel centers. See https://stackoverflow.com/a/37484800/927543
         Vector2 pc(0.5 / img.cols, 0.5 / img.rows);
         bgplane->setUVs(pc, Vector2(pc[0], 1 - pc[1]), Vector2(1 - pc[0], pc[1]), Vector2(1, 1) - pc);
 
@@ -511,22 +795,19 @@ public:
         sceneMgr->getRootSceneNode()->createChildSceneNode()->attachObject(bgplane.get());
     }
 
-    void getScreenshot(OutputArray frame)
+    void getScreenshot(OutputArray frame) CV_OVERRIDE
     {
-        PixelFormat src_type = frameSrc->suggestPixelFormat();
-        int dst_type = src_type == PF_BYTE_RGB ? CV_8UC3 : CV_32FC4;
-
-        frame.create(frameSrc->getHeight(), frameSrc->getWidth(), dst_type);
+        frame.create(rWin->getHeight(), rWin->getWidth(), CV_8UC3);
 
         Mat out = frame.getMat();
-        PixelBox pb(frameSrc->getWidth(), frameSrc->getHeight(), 1, src_type, out.ptr());
-        frameSrc->copyContentsToMemory(pb, pb);
+        PixelBox pb(rWin->getWidth(), rWin->getHeight(), 1, PF_BYTE_RGB, out.ptr());
+        rWin->copyContentsToMemory(pb, pb);
 
         // convert to OpenCV channel order
-        cvtColor(out, out, dst_type == CV_8UC3 ? COLOR_RGB2BGR : COLOR_RGBA2BGRA);
+        cvtColor(out, out, COLOR_RGB2BGR);
     }
 
-    void getDepth(OutputArray depth)
+    void getDepth(OutputArray depth) CV_OVERRIDE
     {
         Camera* cam = sceneMgr->getCamera(title);
         if (!depthRTT)
@@ -534,8 +815,8 @@ public:
             // render into an offscreen texture
             // currently this draws everything twice as OGRE lacks depth texture attachments
             TexturePtr tex = TextureManager::getSingleton().createManual(
-                title + "_Depth", RESOURCEGROUP_NAME, TEX_TYPE_2D, frameSrc->getWidth(),
-                frameSrc->getHeight(), 0, PF_DEPTH, TU_RENDERTARGET);
+                title + "_Depth", RESOURCEGROUP_NAME, TEX_TYPE_2D, rWin->getWidth(),
+                rWin->getHeight(), 0, PF_DEPTH, TU_RENDERTARGET);
             depthRTT = tex->getBuffer()->getRenderTarget();
             depthRTT->addViewport(cam);
             depthRTT->setAutoUpdated(false); // only update when requested
@@ -558,8 +839,15 @@ public:
         ndc = (2 * f * n) / ndc;
     }
 
-    void fixCameraYawAxis(bool useFixed, InputArray _up)
+    void update()
     {
+        rWin->update(false);
+    }
+
+    void fixCameraYawAxis(bool useFixed, InputArray _up) CV_OVERRIDE
+    {
+        if(camman) camman->setFixedYaw(useFixed);
+
         Vector3 up = Vector3::NEGATIVE_UNIT_Y;
         if (!_up.empty())
         {
@@ -569,7 +857,7 @@ public:
         camNode->setFixedYawAxis(useFixed, up);
     }
 
-    void setCameraPose(InputArray tvec, InputArray rot, bool invert)
+    void setCameraPose(InputArray tvec, InputArray rot, bool invert) CV_OVERRIDE
     {
         Quaternion q;
         Vector3 t;
@@ -582,7 +870,7 @@ public:
             camNode->setPosition(t);
     }
 
-    void getCameraPose(OutputArray R, OutputArray tvec, bool invert)
+    void getCameraPose(OutputArray R, OutputArray tvec, bool invert) CV_OVERRIDE
     {
         Matrix3 _R;
         // toOGRE.Inverse() == toOGRE
@@ -611,13 +899,16 @@ public:
         }
     }
 
-    void setCameraIntrinsics(InputArray K, const Size& imsize)
+    void setCameraIntrinsics(InputArray K, const Size& imsize, float zNear, float zFar) CV_OVERRIDE
     {
         Camera* cam = sceneMgr->getCamera(title);
-        _setCameraIntrinsics(cam, K, imsize);
+
+        if(zNear >= 0) cam->setNearClipDistance(zNear);
+        if(zFar >= 0) cam->setFarClipDistance(zFar);
+        if(!K.empty()) _setCameraIntrinsics(cam, K, imsize);
     }
 
-    void setCameraLookAt(const String& target, InputArray offset)
+    void setCameraLookAt(const String& target, InputArray offset) CV_OVERRIDE
     {
         SceneNode* tgt = sceneMgr->getEntity(target)->getParentSceneNode();
 
@@ -630,9 +921,33 @@ public:
 
         camNode->lookAt(tgt->_getDerivedPosition() + _offset, Ogre::Node::TS_WORLD);
     }
+
+	void setEntityLookAt(const String& origin, const String& target, InputArray offset) CV_OVERRIDE
+    {
+		SceneNode* orig = sceneMgr->getEntity(origin)->getParentSceneNode();
+
+		Vector3 _offset = Vector3::ZERO;
+
+        if (!offset.empty())
+        {
+            offset.copyTo(Mat_<Real>(3, 1, _offset.ptr()));
+        }
+
+		if(target.compare("") != 0){
+        SceneNode* tgt = sceneMgr->getEntity(target)->getParentSceneNode();
+		orig->lookAt(tgt->_getDerivedPosition() + _offset, Ogre::Node::TS_WORLD, Ogre::Vector3::UNIT_Z);
+		}else{
+			orig->lookAt(_offset, Ogre::Node::TS_WORLD, Ogre::Vector3::UNIT_Z);
+		}
+    }
+
 };
 
-CV_EXPORTS_W void addResourceLocation(const String& path) { _extraResourceLocations.push_back(path); }
+CV_EXPORTS_W void addResourceLocation(const String& path)
+{
+    CV_Assert(!_app && "must be called before the first createWindow");
+    _extraResourceLocations.insert(Ogre::StringUtil::normalizeFilePath(path, false));
+}
 
 Ptr<WindowScene> createWindow(const String& title, const Size& size, int flags)
 {
@@ -668,8 +983,8 @@ void setMaterialProperty(const String& name, int prop, const Scalar& val)
     CV_Assert(mat);
 
     Pass* rpass = mat->getTechniques()[0]->getPasses()[0];
-    ColourValue col;
 
+    ColourValue col;
     switch (prop)
     {
     case MATERIAL_POINT_SIZE:
@@ -682,10 +997,16 @@ void setMaterialProperty(const String& name, int prop, const Scalar& val)
         rpass->setSceneBlending(SBT_TRANSPARENT_ALPHA);
         rpass->setDepthWriteEnabled(false);
         break;
+    case MATERIAL_DIFFUSE:
+        col = convertColor(val);
+        col.a = rpass->getDiffuse().a;
+        rpass->setDiffuse(col);
+        break;
     case MATERIAL_EMISSIVE:
-        col = ColourValue(val[2], val[1], val[0]) / 255; // BGR as uchar
-        col.saturate();
-        rpass->setEmissive(col);
+        rpass->setEmissive(convertColor(val));
+        break;
+    case MATERIAL_LINE_WIDTH:
+        rpass->setLineWidth(val[0]);
         break;
     default:
         CV_Error(Error::StsBadArg, "invalid or non Scalar property");
@@ -695,20 +1016,23 @@ void setMaterialProperty(const String& name, int prop, const Scalar& val)
 
 void setMaterialProperty(const String& name, int prop, const String& value)
 {
-    CV_Assert(prop == MATERIAL_TEXTURE, _app);
+    CV_Assert_N(prop >= MATERIAL_TEXTURE0, prop <= MATERIAL_TEXTURE3, _app);
 
     MaterialPtr mat = MaterialManager::getSingleton().getByName(name, RESOURCEGROUP_NAME);
     CV_Assert(mat);
 
     Pass* rpass = mat->getTechniques()[0]->getPasses()[0];
 
-    if (rpass->getTextureUnitStates().empty())
+    size_t texUnit = prop - MATERIAL_TEXTURE0;
+    CV_Assert(texUnit <= rpass->getTextureUnitStates().size());
+
+    if (rpass->getTextureUnitStates().size() <= texUnit)
     {
         rpass->createTextureUnitState(value);
         return;
     }
 
-    rpass->getTextureUnitStates()[0]->setTextureName(value);
+    rpass->getTextureUnitStates()[texUnit]->setTextureName(value);
 }
 
 static bool setShaderProperty(const GpuProgramParametersSharedPtr& params, const String& prop,
@@ -765,5 +1089,13 @@ void setMaterialProperty(const String& name, const String& prop, const Scalar& v
     if(!set)
         CV_Error_(Error::StsBadArg, ("shader parameter named '%s' not found", prop.c_str()));
 }
+
+void updateTexture(const String& name, InputArray image)
+{
+    CV_Assert(_app);
+    TexturePtr tex = TextureManager::getSingleton().getByName(name, RESOURCEGROUP_NAME);
+    CV_Assert(tex);
+    _createTexture(name, image.getMat());
+}
 }
 }
